#### Florian Boulant

# Rapport de développement d'une solution au problème des "N reines" utilisant un algorithme génétique

En ce second semestre du master 1 IWOCS, nous avons étudié les différentes
techniques utilisées pour pallier à des problèmes d'optimisation
combinatoire.

Dans le cadre des travaux pratiques de ce module d'enseignement, nous
avons utilisé le problème des "n reines" pour implémenter certaines des
techniques étudiées.

## Présentation du sujet

**Problème des N Reines**

>Placer N reines dans un échiquer de taille N de façon à ce qu'aucune reine
>ne soit en prise. Une reine a une autre pièce en prise lorsqu'elle est
>positionnée sur la même ligne, la même colonne, ou la même diagonale
>que celle-ci.

L'une de ces techniques d'optimisation est l'algorithme génétique, qui vise
à représenter des états de résolution d'un problème sous forme d'individus.
L'algorithme fait évoluer une population d'une taille donnée, avec une
population initiale composée d'individus aléatoires.

L'objectif est d'obtenir un individu d'une qualité optimale, et pour l'atteindre
on génère de nouvelles populations en procédant à une sélection des meilleurs,
qui seront utilisés pour générer une partie de la prochaine génération.
A ces "enfants" on ajoute une partie des meilleurs individus, des pires et de
nouveaux individus aléatoires. On répète cette étape de sélection et croisement
jusqu'à obtenir la solution optimale.

## Choix de procédure de croisement entre deux individus

Le croisement est sans aucun doute l'opération la plus primordiale pour
s'approcher au plus vite de la solution optimale. Afin de ne pas régresser
en croisant les individus, on cherche à conserver des traits bénéfiques dans
l'individu issu d'un croisement, soit l'unicité des colonnes et des lignes.

### Description du croisement

1. Dans un premier temps, on définit un locus dans le premier individu :

Individu 1
| 0 | 1 | 2 | 3 |
|---|---|---|---|
| 3 | 0 | 1 | 2 |
|   | ^ | ^ |   |
Individu 2
| 0 | 1 | 2 | 3 |
|---|---|---|---|
| 1 | 2 | 3 | 0 |
| ^ |   |   | ^ |

2. Dans un second temps, on change l'ordre des colonnes du locus dans le deuxième
   individu pour qu'il corresponde à l'ordre décrit par le locus :

Individu 1
| 0 | 1 | 2 | 3 |
|---|---|---|---|
| 3 | 0 | 1 | 2 |
|   | ^ | ^ |   |
Individu 3
|   0   | 1 | 2 |   3   |
|-------|---|---|-------|
| **0** | 2 | 3 | **1** |
|   ^   |   |   |   ^   |

L'individu résultant de ce croisement est l'enfant des deux croisés.

### Code du croisement dans le projet

```Rust
pub fn nqueens_crossover(a: &Individual, b: &Individual) -> Individual {
    let mut rng: ThreadRng = rand::thread_rng();

    let mut crossover = b.clone();

    //So that locus start isn't on the last cell
    let locus_start = rng.gen_range(0, a.len() - 1);
    let locus_end = rng.gen_range(locus_start + 1, a.len());

    //locus slice of the first individual
    let locus = &a[locus_start..=locus_end];

    let locus_cols = locus.iter().map(|(_, col)| *col).collect::<Vec<usize>>();

    let mut locus_position = 0;
    
    //Goes through each column, and replaces column if it is in the locus.
    for i in 0..crossover.len() {
        let (_, col) = crossover[i];

        if locus_position < locus.len() && locus_cols.contains(&col) {
            crossover[i].1 = locus[locus_position].1;
            locus_position += 1;
        }
    }

    crossover
}
```

## Comparaison des résultats obtenus avec algorithme génétique et recuit simulé

Le verdict est clair : le recuit simulé est plus efficace pour résoudre le
problème des n reines.

J'avais réussi à pousser la dimension de l'échiquier jusqu'aux environs de
300 avec le recuit simulé, alors que je heurte un mur en terme de temps
de recherche à partir d'une taille de 30 avec l'algorithme génétique.

L'une des raisons qui font que j'arrive si vite à un pallier est sûrement
mon choix de croisement. Le croisement est une opération critique et
s'il fournit un mauvais rendement (trop lent par rapport à la plus value
obtenue à l'issue du croisement). Il est fort probable que mon choix de
croisement donne des résultats insuffisants, ce qui ralentit de façon
catastrophique ma recherche de solution, et la coince dans un maximum local.

Ce qui rend l'algorithme du recuit plus efficace que l'algorithme génétique
est le fait que la notion de voisin dans le recuit est plus facilement
modélisable (substitution) que celle du croisement. Passer
de voisin en voisin dans le recuit, plutôt que de travailler avec
un échantillon d'individus qui évolue au fil de la recherche
dans la génétique permet également de ne garder que l'essentiel dans les
étapes derecherche, là où la génétique ne fait que rajouter des étapes
dispensables pour le problème des n reines (sélection, constitution des
prochaines populations...).

En revanche, pour se tirer de minimums locaux, l'algorithme génétique est
potentiellement plus avantageux, car on ne travaille pas que sur un seul
individu mais sur une population, ce qui nous permet de nous sortir de ces
paliers notamment grâce à l'inclusion des pires individus et d'individus
aléatoires dans le générations suivantes. Alors que la seule mesure évitant
d'être coincé dans un minimum local pour le recuit est la température,
qui permet d'ajouter de l'aléatoire dans le choix de solutions voisines,
certes, mais qui diminue également avec le temps, ce qui peut mener à un
coincement.

Mais pour terminer sur cette comparaison, ici les résultats montrent
clairement que le recuit est une bien meilleure approche
(au moins 10x plus efficace) au problème des n reines que
l'algorithme génétique.